from .core import *
from .track import *
from .utils import Log

Log.set_file_prefix("sync")

__all__ = [
    "Check",
    "Config",
    "Index",
    "Pull",
    "Sync",
    "LocalTracks",
    "GithubTracks"
]



# noinspection PyUnresolvedReferences,PyProtectedMember
from argparse import (
    ArgumentParser as ArgumentParserBase,
    RawDescriptionHelpFormatter,
    Action,
    _HelpAction
)
from pathlib import Path
from typing import Optional

from ..__version__ import get_version, get_version_code
from ..core import Index
from ..model import UpdateJson, ModulesJson
from ..utils import GitUtils


class ArgumentParser(ArgumentParserBase):
    def __init__(self, *args, **kwargs):
        if not kwargs.get("formatter_class"):
            kwargs["formatter_class"] = RawDescriptionHelpFormatter
        if "add_help" not in kwargs:
            add_custom_help = True
            kwargs["add_help"] = False
        else:
            add_custom_help = False
        super().__init__(*args, **kwargs)

        if add_custom_help:
            self.add_argument(
                "-h",
                "--help",
                action=_HelpAction,
                help="Show this help message and exit.",
            )


class BoolOrStrAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if values is None or values == "" or values.isspace():
            values = True

        setattr(namespace, self.dest, values)


class Parameters:
    _root_folder: Path
    _github_token: Optional[str]
    _choices: dict

    CONFIG = "config"
    TRACK = "track"
    GITHUB = "github"
    SYNC = "sync"
    INDEX = "index"
    CHECK = "check"
    SITEMAP = "sitemap"

    @classmethod
    def set_root_folder(cls, root: Path):
        cls._root_folder = root

    @classmethod
    def set_github_token(cls, token: Optional[str]):
        cls._github_token = token

    @classmethod
    def print_cmd_help(cls, cmd: Optional[str]):
        cls._choices[cmd].print_help()

    @classmethod
    def generate_parser(cls):
        p = ArgumentParser(
            description="Magisk Modules Repo Util"
        )
        p.add_argument(
            "-v",
            "--version",
            action="version",
            version=get_version(),
            help="Show util version and exit."
        )
        p.add_argument(
            "-V",
            "--version-code",
            action="version",
            version=str(get_version_code()),
            help="Show util version code and exit."
        )

        sub_parsers = p.add_subparsers(
            dest="cmd",
            metavar="command"
        )

        cls.configure_parser_config(sub_parsers)
        cls.configure_parser_track(sub_parsers)
        cls.configure_parser_github(sub_parsers)
        cls.configure_parser_sync(sub_parsers)
        cls.configure_parser_index(sub_parsers)
        cls.configure_parser_check(sub_parsers)
        cls.configure_parser_sitemap(sub_parsers)

        cls._choices = sub_parsers.choices

        return p

    @classmethod
    def configure_parser_config(cls, sub_parsers):
        p = sub_parsers.add_parser(
            cls.CONFIG,
            help="Modify config of repository."
        )
        p.add_argument(
            "-w",
            "--write",
            dest="config_values",
            metavar="KEY=VALUE",
            nargs="+",
            default=None,
            help="Write values to config."
        )
        p.add_argument(
            "--stdin",
            action="store_true",
            help="Write config piped through stdin."
        )
        p.add_argument(
            "--json",
            action="store_true",
            help="Show config of repository."
        )
        p.add_argument(
            "--keys",
            action="store_true",
            help="Show fields available in config."
        )
        p.add_argument(
            "--migrate",
            action="store_true",
            help=f"Migrate config structure and content."
        )

        cls.add_parser_env(p)

    @classmethod
    def configure_parser_track(cls, sub_parsers):
        p = sub_parsers.add_parser(
            cls.TRACK,
            help="Module tracks utility."
        )
        p.add_argument(
            "-l",
            "--list",
            action="store_true",
            help="List tracks in repository."
        )
        p.add_argument(
            "-a",
            "--add",
            dest="track_values",
            metavar="KEY=VALUE",
            nargs="+",
            default=None,
            help="Add a track to repository."
        )
        p.add_argument(
            "-r",
            "--remove",
            dest="remove_module_ids",
            metavar="MODULE_ID",
            nargs="+",
            default=None,
            help="Remove tracks from repository."
        )
        p.add_argument(
            "--stdin",
            action="store_true",
            help="Add track piped through stdin."
        )
        p.add_argument(
            "--keys",
            action="store_true",
            help="Show fields available in track."
        )
        p.add_argument(
            "--migrate",
            action="store_true",
            help=f"Migrate tracks structure and content."
        )

        modify = p.add_argument_group("modify")
        modify.add_argument(
            "-i",
            dest="modify_module_id",
            metavar="MODULE_ID",
            type=str,
            default=None,
            help="Id of the module to modify."
        )
        modify.add_argument(
            "-u",
            "--update",
            dest="update_track_values",
            metavar="KEY=VALUE",
            nargs="+",
            default=None,
            help="Update values to the track."
        )
        modify.add_argument(
            "-d",
            "--remove-key",
            dest="remove_key_list",
            metavar="KEY",
            nargs="+",
            default=None,
            help="Remove keys (and all its values) in the track."
        )
        modify.add_argument(
            "--json",
            action="store_true",
            help="Show the track of module."
        )

        cls.add_parser_env(p)

    @classmethod
    def configure_parser_github(cls, sub_parsers):
        p = sub_parsers.add_parser(
            cls.GITHUB,
            help="Generate tracks from GitHub."
        )
        p.add_argument(
            "-u",
            dest="user_name",
            metavar="USERNAME",
            type=str,
            required=True,
            help="User name or organization name on GitHub."
        )
        p.add_argument(
            "-r",
            dest="repo_names",
            metavar="REPO_NAME",
            nargs="+",
            default=None,
            help="Names of repository, default is all."
        )
        p.add_argument(
            "-d",
            "--date",
            dest="after_date",
            metavar="DATE",
            type=str,
            default="2016-09-08",
            help="Filter out outdated repositories by latest push date, default: {0}.".format("%(default)s")
        )
        p.add_argument(
            "-S",
            "--ssh",
            action="store_true",
            help="Use SSH instead of HTTPS for git clone (deploy SSH key by yourself)."
        )
        p.add_argument(
            "-C",
            "--cover",
            action="store_true",
            help="Overwrite fields of tracks (exclude 'added')."
        )
        p.add_argument(
            "--clear",
            action="store_true",
            help="Remove tracks, exclude those in the current session."
        )

        env = cls.add_parser_env(p, add_quiet=True)
        env.add_argument(
            "--single",
            action="store_true",
            help="Run in single-threaded mode."
        )
        env.add_argument(
            "--token",
            metavar="GITHUB_TOKEN",
            type=str,
            default=cls._github_token,
            help="GitHub REST API Token for PyGitHub, same as 'export GITHUB_TOKEN=...'."
        )

    @classmethod
    def configure_parser_sync(cls, sub_parsers):
        p = sub_parsers.add_parser(
            cls.SYNC,
            help="Sync modules in repository."
        )
        p.add_argument(
            "-i",
            dest="module_ids",
            metavar="MODULE_ID",
            nargs="+",
            default=None,
            help="Ids of modules to update, default is all."
        )
        p.add_argument(
            "-v",
            dest="index_version",
            metavar="VERSION",
            type=int,
            default=Index.latest_version,
            help="Version of the index file ({0}), default: {1}.".format(ModulesJson.filename(), "%(default)s")
        )
        p.add_argument(
            "--diff",
            dest="diff_file",
            metavar="FILE",
            action=BoolOrStrAction,
            nargs="?",
            help="List versions diff of modules."
        )
        p.add_argument(
            "--force",
            action="store_true",
            help="Remove all old versions of modules."
        )
        cls.add_parser_git(p)
        env = cls.add_parser_env(p, add_quiet=True)
        env.add_argument(
            "--single",
            action="store_true",
            help="Run in single-threaded mode."
        )

    @classmethod
    def configure_parser_index(cls, sub_parsers):
        p = sub_parsers.add_parser(
            cls.INDEX,
            help="Generate modules.json from local."
        )
        p.add_argument(
            "-v",
            dest="index_version",
            metavar="VERSION",
            type=int,
            default=Index.latest_version,
            help="Version of the index file ({0}), default: {1}.".format(ModulesJson.filename(), "%(default)s")
        )
        p.add_argument(
            "--json",
            action="store_true",
            help="Show modules.json."
        )
        p.add_argument(
            "--list",
            action="store_true",
            help="List versions of modules."
        )

        cls.add_parser_git(p, add_set_size=False)
        cls.add_parser_env(p)

    @classmethod
    def configure_parser_check(cls, sub_parsers):
        p = sub_parsers.add_parser(
            cls.CHECK,
            help="Content check and migrate."
        )
        p.add_argument(
            "-i",
            dest="module_ids",
            metavar="MODULE_ID",
            nargs="+",
            default=None,
            help="Ids of modules to check, default is all."
        )
        p.add_argument(
            "-I",
            "--id",
            dest="check_id",
            action="store_true",
            help="Check id of the module in all json."
        )
        p.add_argument(
            "-U",
            "--url",
            dest="check_url",
            action="store_true",
            help=f"Check urls of files in {UpdateJson.filename()}."
        )
        p.add_argument(
            "-O",
            "--old",
            dest="remove_old",
            action="store_true",
            help=f"Remove old versions by max_num."
        )

        cls.add_parser_env(p)

    @classmethod
    def configure_parser_sitemap(cls, sub_parsers):
        p = sub_parsers.add_parser(
            cls.SITEMAP,
            help="Sitemap generator."
        )
        p.add_argument(
            "-m",
            "--modules-json",
            dest="modules_json",
            
            help="Path to the modules.json file."
        )
        p.add_argument(
            "-b",
            "--base-url",
            dest="base_url",
            required=True,
            help="Base url that should be used."
        )
        p.add_argument(
            "-o",
            "--output",
            dest="output",
            help="Outputs the file to the given path."
        )

        cls.add_parser_git(p)
        cls.add_parser_env(p)

    @classmethod
    def add_parser_env(cls, p, add_quiet=False):
        env = p.add_argument_group("env")
        env.add_argument(
            "-p",
            "--prefix",
            dest="root_folder",
            metavar="ROOT_FOLDER",
            type=str,
            default=cls._root_folder.as_posix(),
            help="Full path to repository location, current: {0}.".format("%(default)s")
        )

        if add_quiet:
            env.add_argument(
                "-q",
                "--quiet",
                action="store_true",
                help="Show only error logs (piped through stderr)."
            )

        return env

    @classmethod
    def add_parser_git(cls, p, add_set_size=True):
        git = p.add_argument_group("git")
        git.add_argument(
            "--push",
            action="store_true",
            help="Push to git repository."
        )
        git.add_argument(
            "--branch",
            dest="git_branch",
            metavar="GIT_BRANCH",
            type=str,
            default=GitUtils.current_branch(cls._root_folder),
            help="Define branch to push, current: {0}.".format("%(default)s")
        )

        if add_set_size:
            git.add_argument(
                "--size",
                dest="max_size",
                metavar="MAX_SIZE",
                type=float,
                default=50.0,
                help="Filter out oversize zip files, default: {0} MB.".format("%(default)s")
            )

        return git



from .Check import Check
from .Config import Config
from .Index import Index
from .Migrate import Migrate
from .Pull import Pull
from .Sync import Sync

__all__ = [
    "Check",
    "Config",
    "Index",
    "Migrate",
    "Pull",
    "Sync"
]



import functools
import logging
import sys
from datetime import date
from pathlib import Path
from typing import Optional, Dict, Union


class Log:
    _logger_initialized: dict = {}
    _file_prefix: Optional[str] = None
    _enable_stdout: bool = True
    _log_level: int = logging.DEBUG

    def __init__(self, tag: str, *, enable_log: bool = True, log_dir: Optional[Path] = None):
        if log_dir is not None:
            if self._file_prefix is None:
                prefix = tag.lower()
            else:
                prefix = self._file_prefix

            log_file = f"{prefix}_{date.today()}.log"
            log_file = log_dir.joinpath(log_file)
            self.clear(log_dir, prefix)
        else:
            log_file = None

        self._enable_log = enable_log
        self._logging = self.get_logger(
            name=tag,
            log_file=log_file,
            log_level=self._log_level
        )

    def log(self, level: int, msg: str):
        if self._enable_log:
            self._logging.log(level=level, msg=msg)

    def d(self, msg: str):
        self.log(level=logging.DEBUG, msg=msg)

    def i(self, msg: str):
        self.log(level=logging.INFO, msg=msg)

    def w(self, msg: str):
        self.log(level=logging.WARN, msg=msg)

    def e(self, msg: str):
        self.log(level=logging.ERROR, msg=msg)

    @classmethod
    def levels(cls) -> Dict[str, int]:
        return {
            "ERROR": logging.ERROR,
            "WARN": logging.WARNING,
            "WARNING": logging.WARNING,
            "INFO": logging.INFO,
            "DEBUG": logging.DEBUG,
        }

    @classmethod
    def set_file_prefix(cls, name: str):
        cls._file_prefix = name

    @classmethod
    def set_enable_stdout(cls, value: bool):
        cls._enable_stdout = value

    @classmethod
    def set_log_level(cls, level: Union[int, str]):
        levels = cls.levels()

        if isinstance(level, str):
            level = levels.get(level, logging.DEBUG)
        elif isinstance(level, int) and level not in levels.values():
            level = logging.DEBUG

        cls._log_level = level

    @classmethod
    def clear(cls, log_dir: Path, prefix: str, max_num: int = 3):
        log_files = sorted(log_dir.glob(f"{prefix}*"), reverse=True)
        if len(log_files) >= max_num + 1:
            for log_file in log_files[max_num:]:
                log_file.unlink()

    @classmethod
    def get_msg(cls, err: BaseException) -> str:
        msg = "{} " * len(err.args)
        msg = msg.format(*err.args).rstrip()
        return f"{err.__class__.__name__}({msg})"

    @classmethod
    @functools.lru_cache()
    def get_logger(cls, name: str = "root", log_file: Optional[Path] = None, log_level: int = logging.DEBUG):
        logger = logging.getLogger(name)
        if name in cls._logger_initialized:
            return logger
        for logger_name in cls._logger_initialized:
            if name.startswith(logger_name):
                return logger

        formatter = logging.Formatter(
            "[%(asctime)s] %(name)s %(levelname)s: %(message)s",
            datefmt="%Y/%m/%d %H:%M:%S")

        if cls._enable_stdout:
            stdout_handler = logging.StreamHandler(stream=sys.stdout)
            stdout_handler.setFormatter(formatter)
            stdout_handler.addFilter(lambda log: log.levelno < logging.ERROR)
            logger.addHandler(stdout_handler)

        stderr_handler = logging.StreamHandler(stream=sys.stderr)
        stderr_handler.setFormatter(formatter)
        stderr_handler.addFilter(lambda log: log.levelno >= logging.ERROR)
        logger.addHandler(stderr_handler)

        if log_file is not None:
            log_file.parent.mkdir(parents=True, exist_ok=True)

            file_handler = logging.FileHandler(log_file, "a")
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)

        logger.setLevel(log_level)
        cls._logger_initialized[name] = True
        return logger



import shutil

from .Config import Config
from ..model import JsonIO, ConfigJson, TrackJson


class Migrate:
    def __init__(self, root_folder):
        self._root_folder = root_folder

    @staticmethod
    def _config_0_1(old_config):
        return {
            "NAME": old_config.get("repo_name"),
            "BASE_URL": old_config.get("repo_url"),
            "MAX_NUM": old_config.get("max_num"),
            "ENABLE_LOG": old_config.get("show_log"),
            "LOG_DIR": old_config.get("log_dir")
        }

    @staticmethod
    def _config_1_2(old_config):
        return {
            "name": old_config.get("NAME"),
            "base_url": old_config.get("BASE_URL"),
            "max_num": old_config.get("MAX_NUM"),
            "enable_log": old_config.get("ENABLE_LOG"),
            "log_dir": old_config.get("LOG_DIR")
        }

    def config(self):
        config_folder = self._root_folder.joinpath("config")
        json_folder = self._root_folder.joinpath("json")

        config_json_0 = config_folder.joinpath("config.json")
        config_json_1 = json_folder.joinpath("config.json")
        config_json_new = json_folder.joinpath("config.json")

        config = ConfigJson(version=-1)

        # v0
        if config_json_0.exists():
            config.update(JsonIO.load(config_json_0))
            config.version = 0
            shutil.rmtree(config_folder)

        # v1
        if config_json_1.exists():
            config.update(JsonIO.load(config_json_1))

            if config.get("NAME") is not None:
                config.version = 1

        # v0 -> v1
        if config.version == 0:
            config.update(self._config_0_1(config))
            config.version = 1

        # v1 -> v2
        if config.version == 1:
            config.update(self._config_1_2(config))
            config.version = 2

        # write to file
        if config.version != -1:
            config.write(config_json_new)

    def track(self):
        modules_folder = Config.get_modules_folder(self._root_folder)

        for module_folder in modules_folder.glob("*/"):
            json_file = module_folder.joinpath(TrackJson.filename())
            if not json_file.exists():
                continue

            track = TrackJson.load(json_file)

            # remove .disable
            tag_disable = module_folder.joinpath(".disable")
            if tag_disable.exists():
                track.enable = False
                tag_disable.unlink()
            elif track.enable is None:
                track.enable = True

            # write to file (remove null values)
            track.write(json_file)



import os
import sys
from pathlib import Path
from signal import SIGINT, signal

from .cli import Main


# noinspection PyUnresolvedReferences,PyProtectedMember
def signal_handler(*args):
    os._exit(1)

def main():
    signal(SIGINT, signal_handler)
    cwd_folder = Path(__name__).resolve().parent

    try:
        github_token = os.environ["GITHUB_TOKEN"]
    except KeyError:
        github_token = None

    Main.set_default_args(root_folder=cwd_folder, github_token=github_token)
    sys.exit(Main.exec())

if __name__ == "__main__":
    main()


from .AttrDict import AttrDict

class ModuleNote(AttrDict):
    color: str
    title: str
    message: str
    
    @classmethod
    def expected_fields(cls, __type=True):
        if __type:
            return cls.__annotations__

        return {k: v.__name__ for k, v in cls.__annotations__.items()}


from typing import Optional, Any, Callable


class Result:
    def __init__(self, value: Optional[Any] = None, error: Optional[BaseException] = None):
        self.value = value
        self.error = error

    @property
    def is_success(self):
        return self.error is None

    @property
    def is_failure(self):
        return not self.is_success

    def get_or_default(self, default: Any) -> Any:
        return self.value or default

    @classmethod
    def catching(cls):
        def decorator(func: Callable[..., Any]) -> Callable[..., Result]:
            def wrapper(*args, **kwargs):
                try:
                    value = func(*args, **kwargs)
                    return Result(value=value)
                except BaseException as err:
                    return Result(error=err)

            return wrapper
        return decorator



import json
import yaml
from pathlib import Path

from .AttrDict import AttrDict
from .ArchiveIO import ArchiveIO
from ..error import MagiskModuleError

from .JsonIO import JsonIO

from .ModuleNote import ModuleNote
from .ModuleFeatures import ModuleFeatures
from .ModuleManager import ModuleManager
from .RootSolutions import RootSolutions

class LocalModule(AttrDict):
    id: str
    name: str
    version: str
    versionCode: int
    author: str
    description: str
    
    added: float
    timestamp: float
    size: float
    
    # FoxMMM supported props
    maxApi: int
    minApi: int
    
    # MMRL supported props
    category: str
    categories: list[str]
    icon: str
    homepage: str
    donate: str
    support: str
    cover: str
    screenshots: list[str]
    license: str
    screenshots: list[str]
    readme: str
    require: list[str]
    verified: bool
    note: ModuleNote
    features: ModuleFeatures
    root: RootSolutions
    manager: ModuleManager

    @classmethod
    def clean_json(cls, data):
        if isinstance(data, dict):
            cleaned_dict = {
                key: cls.clean_json(value)
                for key, value in data.items()
                if value not in (None, [], {}, "")
            }
            return {k: v for k, v in cleaned_dict.items() if v not in (None, [], {}, "")}
        elif isinstance(data, list):
            cleaned_list = [cls.clean_json(item) for item in data]
            return [item for item in cleaned_list if item not in (None, [], {}, "")]
        return data

    @classmethod
    def load(cls, file, track, config):    
        zip_compression = track.deep_get("options.archive.compression", default="stored")
        disable_metadata = track.deep_get("options.disableRemoteMetadata", default=False)
        
        cls._zipfile = ArchiveIO(file=file, mode="r", compression=zip_compression)
        fields = cls.expected_fields()

        cleaned_track = cls.clean_json(track)

        try:
            if ("#MAGISK" not in cls._zipfile.file_read("META-INF/com/google/android/updater-script")):
                raise
            if (not cls._zipfile.file_exists("META-INF/com/google/android/update-binary")):
                raise
        except BaseException:
            msg = f"{file.name} is not a magisk module"
            raise MagiskModuleError(msg)

        try:
            props = cls._zipfile.file_read("module.prop")
        except BaseException as err:
            raise MagiskModuleError(err.args)

        obj = AttrDict()
        for item in props.splitlines():
            prop = item.split("=", maxsplit=1)
            if len(prop) != 2:
                continue

            key, value = prop
            if key == "" or key.startswith("#") or key not in fields:
                continue

            _type = fields[key]
            obj[key] = _type(value)

        local_module = LocalModule()
        for key in fields.keys():
            if config.allowedCategories and key == "categories" and cleaned_track.get("categories"):
                local_module[key] = JsonIO.filterArray(config.allowedCategories, cleaned_track.get(key))
            else:
                value = cleaned_track.get(key) if cleaned_track.get(key) is not None else obj.get(key)
                if value is not None and value is not False:
                    local_module[key] = value

        try:
            if not disable_metadata:
                has_yaml = cls._zipfile.file_exists("common/repo.yaml")
                if has_yaml:
                    raw_json = yaml.load(cls._zipfile.file_read("common/repo.yaml"), Loader=yaml.FullLoader)
                else:
                    raw_json = json.loads(cls._zipfile.file_read("common/repo.json"))
                raw_json = cls.clean_json(raw_json)

                for item in raw_json.items():
                    key, value = item

                    _type = fields[key]
                    obj[key] = _type(value)

                for key in fields.keys():
                    value = obj.get(key)
                    if value is not None and value is not False: 
                        local_module[key] = value
                    
        except BaseException:
            pass

        local_module.verified = track.verified or False
        local_module.added = track.added or 0
        local_module.timestamp = track.last_update
        local_module.size = Path(file).stat().st_size        
        
        features = {
            "service": cls._zipfile.file_exists(f"service.sh") or cls._zipfile.file_exists(f"common/service.sh"),
            "post_fs_data": cls._zipfile.file_exists(f"post-fs-data.sh") or cls._zipfile.file_exists(f"common/post-fs-data.sh"),
            # system.prop
            "resetprop": cls._zipfile.file_exists(f"system.prop") or cls._zipfile.file_exists(f"common/system.prop"),
            "sepolicy": cls._zipfile.file_exists(f"sepolicy.rule"),
            
            "zygisk": cls._zipfile.file_exists(f"zygisk/"),
            "action": cls._zipfile.file_exists(f"action.sh") or cls._zipfile.file_exists(f"common/action.sh"),
            
            # KernelSU
            "webroot": cls._zipfile.file_exists(f"webroot/index.html"),
            "post_mount": cls._zipfile.file_exists(f"post-mount.sh") or cls._zipfile.file_exists(f"common/post-mount.sh"),
            "boot_completed": cls._zipfile.file_exists(f"boot-completed.sh") or cls._zipfile.file_exists(f"common/boot-completed.sh"),

            # MMRL
            "modconf": cls._zipfile.file_exists(f"system/usr/share/mmrl/config/{local_module.id}/index.jsx"),
            
            "apks": len([name for name in cls._zipfile.namelist() if name.endswith('.apk')]) != 0
        }
        
        local_module.features = {k: v for k, v in features.items() if v is not None and v is not False}

        return cls.clean_json(local_module)
   
    @classmethod
    def expected_fields(cls, __type=True):
        if __type:
            return cls.__annotations__

        return {k: v.__name__ for k, v in cls.__annotations__.items() if v is not None and v is not False}



from datetime import datetime
from typing import Optional, List

import requests
from dateutil.parser import parse


class GitHubGraphQLAPI:
    def __init__(self, api_token: str):
        self._api_token = api_token

    def _graphql_query(self, query: str) -> Optional[dict]:
        query = {"query": query}

        response = requests.post(
            url="https://api.github.com/graphql",
            headers={
                "Authorization": f"bearer {self._api_token}",
                "Content-Type": "application/json",
            },
            json=query
        )

        if response.ok:
            return response.json()
        else:
            return None

    def _query_repository(self, owner: str, name: str, query: str) -> Optional[dict]:
        params = "owner: \"{}\", name: \"{}\"".format(owner, name)
        _query = "query { repository(%s) { %s } }" % (params, query)
        result = self._graphql_query(_query)

        try:
            data = result.get("data")
            repository = data.get("repository")
            return repository
        except AttributeError:
            return None

    def get_sponsor_url(self, owner: str, name: str) -> List[str]:
        repository = self._query_repository(
            owner=owner,
            name=name,
            query="fundingLinks { platform url }"
        )
        if repository is None:
            return list()

        links = list()
        funding_links = repository["fundingLinks"]

        for item in funding_links:
            if item["platform"] == "GITHUB":
                name = item["url"].split("/")[-1]
                links.append(f"https://github.com/sponsors/{name}")
            else:
                links.append(item["url"])

        return links

    def get_homepage_url(self, owner: str, name: str) -> Optional[str]:
        repository = self._query_repository(
            owner=owner,
            name=name,
            query="homepageUrl"
        )
        if repository is None:
            return None

        homepage_url = repository["homepageUrl"]
        if homepage_url != "":
            return homepage_url
        else:
            return None

    def get_pushed_at(self, owner: str, name: str) -> Optional[datetime]:
        repository = self._query_repository(
            owner=owner,
            name=name,
            query="pushedAt"
        )
        if repository is None:
            return None

        try:
            return parse(repository["pushedAt"])
        except TypeError:
            return None



from functools import reduce

class AttrDict(dict):
    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __getattr__(self, item):
        return self.get(item)

    def __hash__(self):
        return hash(frozenset(self.__dict__.items()))

    def deep_get(self, keys, default=None):
        return reduce(lambda d, key: d.get(key, default) if isinstance(d, dict) else default, keys.split("."), self)

    def copy(self, **kwargs):
        new = super().copy()
        new.update(**kwargs)

        return AttrDict(new)



import os
import magic

from pathlib import Path

from ..error import ConfigError
from ..model import ConfigJson, JsonIO
from ..utils import Log, StrUtils


class Config(ConfigJson):
    def __init__(self, root_folder):
        self._log = Log("Config", enable_log=True)
        self._root_folder = root_folder
        self._mime = magic.Magic(mime=True)

        json_file = self.json_folder.joinpath(ConfigJson.filename())
        if not json_file.exists():
            raise FileNotFoundError(json_file.as_posix())

        obj = JsonIO.load(json_file)
        super().__init__(obj)

        self._check_values()

        self._log = Log("Config", enable_log=self.enable_log, log_dir=self.log_dir)
        for key in ConfigJson.expected_fields().keys():
            self._log.d(f"{key} = {self.get(key)}")

    def _check_values(self):
        default = self.default()

        name = self.get("name", default.name)
        if name == default.name:
            self._log.w("_check_values: 'name' is undefined")

        base_url = self.get("base_url", default.base_url)
        if base_url == default.base_url:
            raise ConfigError("'base_url' is undefined")
        elif not StrUtils.is_with(base_url, "https", "/"):
            raise ConfigError("'base_url' must start with 'https' and end with '/'")

        max_num = self.get("max_num", default.max_num)
        enable_log = self.get("enable_log", default.enable_log)

        website = self.get("website", default.website)
        support = self.get("support", default.support)
        donate = self.get("donate", default.donate)
        submission = self.get("submission", default.submission)
        description = self.get("description", default.description)

        log_dir = self.get("log_dir", default.log_dir)
        if log_dir != default.log_dir:
            log_dir = Path(log_dir)

            if not log_dir.is_absolute():
                log_dir = self._root_folder.joinpath(log_dir)

        self.update(
            name=name,
            website=website,
            support=support,
            donate=donate,
            submission=submission,
            description=description,
            cover=self.get_cover,
            base_url=base_url,
            max_num=max_num,
            enable_log=enable_log,
            log_dir=log_dir
        )


    @property
    def get_cover(self):
        cover_file = self.assets_folder.joinpath("cover.webp")
        cover = None

        if cover_file.exists() and cover_file.suffix.lower() == '.webp':
            cover_mime_type = self._mime.from_file(cover_file)

            if cover_mime_type.lower() == 'image/webp':
                cover = f"{self.base_url}assets/cover.webp"
            else:
                self._log.warning(f"get_cover: '{cover_file.name}' is not a valid WebP image.")
        else:
            self._log.info(f"get_cover: '{cover_file.name}' does not exist or is not a WebP file.")

        return cover

    @property
    def json_folder(self):
        return self.get_json_folder(self._root_folder)

    @property
    def assets_folder(self):
        return self.get_assets_folder(self._root_folder)

    @property
    def modules_folder(self):
        return self.get_modules_folder(self._root_folder)

    @property
    def local_folder(self):
        return self.get_local_folder(self._root_folder)

    @classmethod
    def get_json_folder(cls, root_folder):
        return root_folder.joinpath("json")

    @classmethod
    def get_assets_folder(cls, root_folder):
        return root_folder.joinpath("assets")

    @classmethod
    def get_modules_folder(cls, root_folder):
        return root_folder.joinpath("modules")

    @classmethod
    def get_local_folder(cls, root_folder):
        return root_folder.joinpath("local")



from .AttrDict import AttrDict

class ModuleManagerSolution(AttrDict):
    min: int
    devices: list[str]
    arch: list[str]
    require: list[str]
    
    @classmethod
    def expected_fields(cls, __type=True):
        if __type:
            return cls.__annotations__

        return {k: v.__name__ for k, v in cls.__annotations__.items() if v is not None and v is not False}

class ModuleManager(AttrDict):
    magisk: ModuleManagerSolution
    kernelsu: ModuleManagerSolution
    ksunext: ModuleManagerSolution
    apatch: ModuleManagerSolution
    
    @classmethod
    def expected_fields(cls, __type=True):
        if __type:
            return cls.__annotations__

        return {k: v.__name__ for k, v in cls.__annotations__.items() if v is not None and v is not False}
    


from .AttrDict import AttrDict

class TrackOptionsArchive(AttrDict):
    compression: str

    @classmethod
    def expected_fields(cls, __type=True):
        if __type:
            return cls.__annotations__

        return {k: v.__name__ for k, v in cls.__annotations__.items()}

class TrackOptions(AttrDict):
    archive: TrackOptionsArchive
    disableRemoteMetadata: bool
    
    @classmethod
    def expected_fields(cls, __type=True):
        if __type:
            return cls.__annotations__

        return {k: v.__name__ for k, v in cls.__annotations__.items()}


import shutil

from .Config import Config
from ..error import Result
from ..model import (
    LocalModule,
    AttrDict,
    MagiskUpdateJson,
    OnlineModule,
    TrackType,
    UpdateJson
)
from ..utils import (
    Log,
    HttpUtils,
    GitUtils,
    StrUtils
)


class Pull:
    _max_size = 50

    def __init__(self, root_folder, config):
        self._log = Log("Pull", enable_log=config.enable_log, log_dir=config.log_dir)

        self._local_folder = Config.get_local_folder(root_folder)
        self._modules_folder = Config.get_modules_folder(root_folder)
        self._config = config

    @staticmethod
    def _copy_file(old, new, delete_old):
        shutil.copy(old, new)
        if delete_old:
            old.unlink()

    @staticmethod
    @Result.catching()
    def _download(url, out):
        return HttpUtils.download(url, out)

    def _check_changelog(self, module_id, file):
        text = file.read_text()
        if StrUtils.is_html(text):
            self._log.w(f"_check_changelog: [{module_id}] -> unsupported type (html text)")
            return False
        else:
            return True

    def _check_version_code(self, module_id, version_code):
        module_folder = self._modules_folder.joinpath(module_id)
        json_file = module_folder.joinpath(UpdateJson.filename())

        if not json_file.exists():
            return True

        update_json = UpdateJson.load(json_file)
        if version_code is None:
          self._log.e(f"_check_version_code: [{module_id}] has no version code set")
          return False
        if len(update_json.versions) != 0 and version_code > update_json.versions[-1].versionCode:
            return True

        self._log.i(f"_check_version_code: [{module_id}] -> already the latest version")
        return False

    def _get_file_url(self, module_id, file):
        module_folder = self._modules_folder.joinpath(module_id)
        url = f"{self._config.base_url}{self._modules_folder.name}/{module_id}/{file.name}"

        if not (file.parent == module_folder and file.exists()):
            raise FileNotFoundError(f"{file} is not in {module_folder}")
        else:
            return url

    def _get_changelog_common(self, module_id, changelog):
        if changelog is None:
            return None
        elif isinstance(changelog, str) and changelog == "":
            return None

        if StrUtils.is_url(changelog):
            if StrUtils.is_blob_url(changelog):
                msg = f"'{changelog}' is not unsupported type, please use 'https://raw.githubusercontent.com'"
                self._log.w(f"_get_changelog_common: [{module_id}] -> {msg}")
                return None

            changelog_file = self._modules_folder.joinpath(module_id, f"{module_id}.md")
            result = self._download(changelog, changelog_file)
            if result.is_failure:
                msg = Log.get_msg(result.error)
                self._log.e(f"_get_changelog_common: [{module_id}] -> {msg}")
                changelog_file = None

        else:
            changelog_file = self._modules_folder.joinpath(module_id, f"{module_id}.md")
            changelog_file.write_text(changelog)

        if changelog_file is not None:
            if not self._check_changelog(module_id, changelog_file):
                changelog_file.unlink()
                changelog_file = None

        return changelog_file

    def _from_zip_common(self, track, zip_file, changelog_file, *, delete_tmp):
        module_folder = self._modules_folder.joinpath(track.id)

        def remove_file():
            if delete_tmp:
                zip_file.unlink()
            if delete_tmp and changelog_file is not None:
                changelog_file.unlink()

        zip_file_size = zip_file.stat().st_size / (1024 ** 2)
        if zip_file_size > self._max_size:
            new_module_folder = self._local_folder.joinpath(track.id)
            msg = f"zip file is oversize ({self._max_size} MB), move this module to {new_module_folder}"
            self._log.w(f"_from_zip_common: [{track.id}] -> {msg}")
            shutil.rmtree(new_module_folder, ignore_errors=True)
            shutil.move(module_folder, new_module_folder)

            return None

        @Result.catching()
        def get_online_module():
            local_module = LocalModule.load(zip_file, track, self._config)
            return OnlineModule.from_dict(local_module)

        result = get_online_module()
        if result.is_failure:
            msg = Log.get_msg(result.error)
            self._log.e(f"_from_zip_common: [{track.id}] -> {msg}")
            remove_file()
            return None
        else:
            online_module: OnlineModule = result.value

        target_zip_file = module_folder.joinpath(online_module.zipfile_name)
        if self._check_version_code(track.id, online_module.versionCode):
            self._copy_file(zip_file, target_zip_file, delete_tmp)
        else:
            remove_file()
            return None

        target_changelog_file = module_folder.joinpath(online_module.changelog_filename)
        changelog_url = ""
        if changelog_file is not None:
            self._copy_file(changelog_file, target_changelog_file, delete_tmp)
            changelog_url = self._get_file_url(track.id, target_changelog_file)

        # For OnlineModule.to_VersionItem
        online_module.latest = AttrDict(
            zipUrl=self._get_file_url(track.id, target_zip_file),
            size=target_zip_file.stat().st_size, 
            changelog=changelog_url
        )

        return online_module

    def from_json(self, track, *, local):
        if local:
            update_to = self._local_folder.joinpath(track.update_to)
        else:
            update_to = track.update_to

        @Result.catching()
        def load_json():
            return MagiskUpdateJson.load(update_to)

        result = load_json()
        if result.is_failure:
            msg = Log.get_msg(result.error)
            self._log.e(f"from_json: [{track.id}] -> {msg}")
            return None, 0.0
        else:
            update_json: MagiskUpdateJson = result.value

        if not self._check_version_code(track.id, update_json.versionCode):
            return None, 0.0

        zip_file = self._modules_folder.joinpath(track.id, f"{track.id}.zip")

        result = self._download(update_json.zipUrl, zip_file)
        if result.is_failure:
            msg = Log.get_msg(result.error)
            self._log.e(f"from_json: [{track.id}] -> {msg}")
            return None, 0.0
        else:
            last_modified = result.value

        changelog = self._get_changelog_common(track.id, update_json.changelog)
        online_module = self._from_zip_common(track, zip_file, changelog, delete_tmp=True)
        return online_module, last_modified

    def from_url(self, track):
        zip_file = self._modules_folder.joinpath(track.id, f"{track.id}.zip")

        result = self._download(track.update_to, zip_file)
        if result.is_failure:
            msg = Log.get_msg(result.error)
            self._log.e(f"from_url: [{track.id}] -> {msg}")
            return None, 0.0
        else:
            last_modified = result.value

        changelog = self._get_changelog_common(track.id, track.changelog)
        online_module = self._from_zip_common(track, zip_file, changelog, delete_tmp=True)
        return online_module, last_modified

    def from_git(self, track):
        zip_file = self._modules_folder.joinpath(track.id, f"{track.id}.zip")

        @Result.catching()
        def git_clone():
            return GitUtils.clone_and_zip(track.update_to, zip_file)

        result = git_clone()
        if result.is_failure:
            msg = Log.get_msg(result.error)
            self._log.e(f"from_git: [{track.id}] -> {msg}")
            return None, 0.0
        else:
            last_committed = result.value

        changelog = self._get_changelog_common(track.id, track.changelog)
        online_module = self._from_zip_common(track, zip_file, changelog, delete_tmp=True)
        return online_module, last_committed

    def from_zip(self, track):
        zip_file = self._local_folder.joinpath(track.update_to)
        changelog = self._local_folder.joinpath(track.changelog)
        last_modified = zip_file.stat().st_mtime

        if not zip_file.exists():
            msg = f"{track.update_to} is not in {self._local_folder}"
            self._log.i(f"from_zip: [{track.id}] -> {msg}")
            return None, 0.0

        if not changelog.exists():
            changelog = None

        online_module = self._from_zip_common(track, zip_file, changelog, delete_tmp=False)
        return online_module, last_modified

    def from_track(self, track):
        self._log.d(f"from_track: [{track.id}] -> type: {track.type.name}")

        if track.type == TrackType.ONLINE_JSON:
            return self.from_json(track, local=False)
        elif track.type == TrackType.ONLINE_ZIP:
            return self.from_url(track)
        elif track.type == TrackType.GIT:
            return self.from_git(track)
        elif track.type == TrackType.LOCAL_JSON:
            return self.from_json(track, local=True)
        elif track.type == TrackType.LOCAL_ZIP:
            return self.from_zip(track)

        self._log.e(f"from_track: [{track.id}] -> unsupported type [{track.update_to}]")
        return None, 0.0

    @classmethod
    def set_max_size(cls, value):
        cls._max_size = value



import shutil

from .Config import Config
from .Index import Index
from .Pull import Pull
from ..model import TrackJson, UpdateJson, VersionItem
from ..track import LocalTracks
from ..utils import Log


class Check:
    def __init__(self, root_folder, config):
        self._log = Log("Check", enable_log=config.enable_log, log_dir=config.log_dir)

        self._local_folder = Config.get_local_folder(root_folder)
        self._modules_folder = Config.get_modules_folder(root_folder)
        self._tracks = LocalTracks(self._modules_folder, config)
        self._config = config

    def _get_file_url(self, module_id, file):
        func = getattr(Pull, "_get_file_url")
        return func(self, module_id, file)

    def _get_tracks(self, module_ids, new):
        if new or self._tracks.size == 0:
            return self._tracks.get_tracks(module_ids)

        return self._tracks.tracks

    def _check_folder(self, track, target_id):
        if track.id == target_id:
            return True

        msg = f"id is not same as in module.prop ({target_id})"
        self._log.d(f"_check_folder: [{track.id}] -> {msg}")

        old_module_folder = self._modules_folder.joinpath(track.id)
        new_module_folder = self._modules_folder.joinpath(target_id)

        if new_module_folder.exists():
            new_module_folder = self._local_folder.joinpath(track.id)
            msg = f"{target_id} already exists, move the old to {new_module_folder.as_posix()}"
            self._log.w(f"_check_folder: [{track.id}] -> {msg}")
            shutil.move(old_module_folder, new_module_folder)

            return True

        old_module_folder.rename(new_module_folder)
        track.update(id=target_id)

        return False

    def _get_new_version_item(self, track, item):
        module_folder = self._modules_folder.joinpath(track.id)

        zipfile_name = item.zipfile_name
        zipfile = module_folder.joinpath(zipfile_name)
        if not zipfile.exists():
            msg = f"{zipfile_name} does not exist, it will be removed from {UpdateJson.filename()}"
            self._log.w(f"_get_new_version_item: [{track.id}] -> {msg}")
            return None

        new_zip_url = self._get_file_url(track.id, zipfile)

        changelog = module_folder.joinpath(item.changelog_filename)
        new_changelog_url = ""
        if changelog.exists() and changelog.is_file():
            new_changelog_url = self._get_file_url(track.id, changelog)

        return VersionItem(
            timestamp=item.timestamp,
            version=item.version,
            versionCode=item.versionCode,
            zipUrl=new_zip_url,
            changelog=new_changelog_url,
            size=zipfile.stat().st_size
        )

    def _check_update_json(self, track, update_json, check_id):
        new_update_json = UpdateJson(
            id=track.id,
            timestamp=update_json.timestamp,
            versions=list()
        )

        for item in update_json.versions:
            if check_id and item.id == track.id:
                continue
            elif not check_id and item.zipUrl.startswith(self._config.base_url):
                continue

            new_item = self._get_new_version_item(track, item)
            if new_item is None:
                continue

            new_update_json.versions.append(new_item)

        if len(new_update_json.versions) != 0:
            update_json.clear()
            update_json.update(new_update_json)
            return False

        return True

    def get_online_module(self, module_id, zip_file):
        func = getattr(Index, "get_online_module")
        return func(self, module_id, zip_file)

    def url(self, module_ids=None, new=False):
        for track in self._get_tracks(module_ids, new):
            module_folder = self._modules_folder.joinpath(track.id)
            update_json_file = module_folder.joinpath(UpdateJson.filename())
            if not update_json_file.exists():
                continue

            update_json = UpdateJson.load(update_json_file)

            if not self._check_update_json(track, update_json, False):
                self._log.i(f"url: [{track.id}] -> {UpdateJson.filename()} has been updated")
                update_json.write(update_json_file)

    def ids(self, module_ids=None, new=False):
        for track in self._get_tracks(module_ids, new):
            old_id = track.id
            module_folder = self._modules_folder.joinpath(old_id)

            zip_files = sorted(
                module_folder.glob("*.zip"),
                key=lambda f: f.stat().st_mtime,
                reverse=True
            )

            if len(zip_files) == 0:
                continue

            latest_zip = zip_files[0]
            online_module = self.get_online_module(track.id, latest_zip)
            if online_module is None:
                continue

            if not self._check_folder(track, online_module.id):
                self._log.i(f"ids: [{old_id}] -> track has been migrated to {track.id}")
                module_folder = self._modules_folder.joinpath(track.id)
                
                has_yaml_track = module_folder.joinpath("track.yaml").exists()
                track_json_file = module_folder.joinpath(TrackJson.filename(has_yaml_track))
                
                track.write(track_json_file)

            update_json_file = module_folder.joinpath(UpdateJson.filename())
            if not update_json_file.exists():
                continue

            update_json = UpdateJson.load(update_json_file)
            if not self._check_update_json(track, update_json, True):
                self._log.i(f"ids: [{track.id}] -> {UpdateJson.filename()} has been updated")
                update_json.write(update_json_file)

    def old(self, module_ids=None, new=False):
        for track in self._get_tracks(module_ids, new):
            module_folder = self._modules_folder.joinpath(track.id)
            update_json_file = module_folder.joinpath(UpdateJson.filename())
            if not update_json_file.exists():
                continue

            update_json = UpdateJson.load(update_json_file)

            max_num = self._config.max_num
            if track.max_num is not None:
                max_num = track.max_num

            if len(update_json.versions) <= max_num:
                continue

            old_versions = update_json.versions[:-max_num]
            for old_item in old_versions:
                update_json.versions.remove(old_item)
                zipfile = module_folder.joinpath(old_item.zipfile_name)
                changelog = module_folder.joinpath(old_item.changelog_filename)

                for path in [zipfile, changelog]:
                    if not (path.exists() and path.is_file()):
                        continue

                    self._log.d(f"old: [{track.id}] -> remove {path.name}")
                    path.unlink()

            self._log.i(f"old: [{track.id}] -> {UpdateJson.filename()} has been updated")
            update_json.write(update_json_file)

            self._log.i(f"old: [{track.id}] -> {TrackJson.filename()} has been updated")
            track_json_file = module_folder.joinpath(TrackJson.filename())
            track.versions = len(update_json.versions)
            track.write(track_json_file)



from .BaseTracks import BaseTracks
from .LocalTracks import LocalTracks

try:
    from .GithubTracks import GithubTracks
except ImportError:
    from .BaseTracks import BaseTracks as GithubTracks


__all__ = [
    "BaseTracks",
    "GithubTracks",
    "LocalTracks"
]



from .AttrDict import AttrDict

class ModuleFeatures(AttrDict):
    service: bool
    post_fs_data: bool
    # system.prop
    resetprop: bool
    sepolicy: bool
    zygisk: bool
    apks: bool
    
    # KernelSU
    webroot: bool
    post_mount: bool
    boot_completed: bool
    
    # MMRL
    modconf: bool
    
    @classmethod
    def expected_fields(cls, __type=True):
        if __type:
            return cls.__annotations__

        return {k: v.__name__ for k, v in cls.__annotations__.items()}


import json
import logging
import os
import sys
import json
import xml.etree.ElementTree as ET
from argparse import Namespace
from pathlib import Path
from typing import Sequence, Type, Tuple
from xml.dom import minidom
from datetime import datetime
from pathlib import Path

from dateutil.parser import parse

from .Parameters import Parameters
from ..core import (
    Check,
    Config,
    Index,
    Migrate,
    Pull,
    Sync
)
from ..model import TrackJson, JsonIO, ConfigJson, AttrDict
from ..track import LocalTracks, GithubTracks
from ..utils import (
    Log,
    Sitemap
)


class SafeArgs(Namespace):
    def __init__(self, args: Namespace):
        super().__init__(**args.__dict__)

    def __getattr__(self, item):
        if item not in self.__dict__:
            return None

        return self.__dict__[item]


class Main:
    _args: SafeArgs
    CODE_FAILURE = 1
    CODE_SUCCESS = 0

    @classmethod
    def set_default_args(cls, **kwargs):
        root_folder = kwargs.get("root_folder", os.getcwd())
        root_folder = Path(root_folder).resolve()
        Parameters.set_root_folder(root_folder)

        github_token = kwargs.get("github_token")
        Parameters.set_github_token(github_token)

    @classmethod
    def exec(cls) -> int:
        parser = Parameters.generate_parser()
        cls._args = SafeArgs(parser.parse_args())

        code = cls._check_args()
        if code == cls.CODE_FAILURE:
            if cls._args.cmd is None:
                parser.print_help()
            else:
                Parameters.print_cmd_help(cls._args.cmd)

        return code

    @classmethod
    def _check_args(cls) -> int:
        if cls._args.cmd is None:
            return cls.CODE_FAILURE
        elif cls._args.cmd == Parameters.CONFIG:
            return cls.config()
        elif cls._args.cmd == Parameters.TRACK:
            return cls.track()
        elif cls._args.cmd == Parameters.GITHUB:
            return cls.github()
        elif cls._args.cmd == Parameters.SYNC:
            return cls.sync()
        elif cls._args.cmd == Parameters.INDEX:
            return cls.index()
        elif cls._args.cmd == Parameters.CHECK:
            return cls.check()
        elif cls._args.cmd == Parameters.SITEMAP:
            return cls.sitemap()

    @classmethod
    def config(cls) -> int:
        root_folder = Path(cls._args.root_folder).resolve()
        json_folder = Config.get_json_folder(root_folder)
        json_file = json_folder.joinpath(Config.filename())

        if cls._args.migrate:
            migrate = Migrate(root_folder)
            migrate.config()
            return cls.CODE_SUCCESS

        if cls._args.config_values is not None:
            _dict, _error = json_parse(cls._args.config_values, ConfigJson)
            if len(_error) != 0:
                error = json.dumps(obj=_error, indent=2)
                print_error(error)

            else:
                if json_file.exists():
                    config = ConfigJson.load(json_file)
                else:
                    config = ConfigJson()

                config.update(_dict)
                config.write(json_file)

        elif cls._args.stdin:
            config_dict = json.load(fp=sys.stdin)
            ConfigJson.write(config_dict, json_file)

        elif cls._args.json and json_file.exists():
            config_dict = JsonIO.load(json_file)
            print_json(config_dict)

        elif cls._args.keys:
            fields = ConfigJson.expected_fields(False)
            print_json(fields)

        else:
            return cls.CODE_FAILURE

        return cls.CODE_SUCCESS

    @classmethod
    def track(cls) -> int:
        root_folder = Path(cls._args.root_folder).resolve()
        modules_folder = Config.get_modules_folder(root_folder)
        Log.set_enable_stdout(False)

        if cls._args.migrate:
            migrate = Migrate(root_folder)
            migrate.track()
            return cls.CODE_SUCCESS

        if cls._args.list:
            config = Config(root_folder)
            tracks = LocalTracks(modules_folder=modules_folder, config=config)
            markdown_text = tracks.get_tracks_table()
            print(markdown_text)

        elif cls._args.track_values is not None:
            _dict, _error = json_parse(cls._args.track_values, TrackJson)
            if len(_error) != 0:
                error = json.dumps(obj=_error, indent=2)
                print_error(error)

            else:
                track = TrackJson(_dict)
                LocalTracks.add_track(
                    track=track,
                    modules_folder=modules_folder,
                    cover=True
                )

        elif cls._args.remove_module_ids is not None:
            for module_id in cls._args.remove_module_ids:
                LocalTracks.del_track(
                    module_id=module_id,
                    modules_folder=modules_folder
                )

        elif cls._args.stdin:
            track = TrackJson(json.load(fp=sys.stdin))
            module_folder = modules_folder.joinpath(track.id)

            has_yaml_track = module_folder.joinpath("track.yaml").exists()
            json_file = module_folder.joinpath(TrackJson.filename(has_yaml_track))
            
            track.write(json_file)

        elif cls._args.keys:
            keys = TrackJson.expected_fields(False)
            print_json(keys)

        elif cls._args.modify_module_id is not None:
            module_folder = modules_folder.joinpath(cls._args.modify_module_id)
            
            has_yaml_track = module_folder.joinpath("track.yaml").exists()
            json_file = module_folder.joinpath(TrackJson.filename(has_yaml_track))

            if not json_file.exists():
                print_error(f"There is no track for this id ({cls._args.modify_module_id})")
                return cls.CODE_SUCCESS

            if cls._args.update_track_values is not None:
                _dict, _error = json_parse(cls._args.update_track_values, TrackJson)
                if len(_error) != 0:
                    error = json.dumps(obj=_error, indent=2)
                    print_error(error)

                else:
                    track = TrackJson(_dict)
                    track.update(id=cls._args.modify_module_id)
                    LocalTracks.update_track(
                        track=track,
                        modules_folder=modules_folder
                    )

            elif cls._args.remove_key_list is not None and json_file.exists():
                track = TrackJson.load(json_file)
                for key in cls._args.remove_key_list:
                    track.pop(key, None)
                track.write(json_file)

            elif cls._args.json and json_file.exists():
                track = TrackJson.load(json_file)
                print_json(track)

            else:
                return cls.CODE_FAILURE

        else:
            return cls.CODE_FAILURE

        return cls.CODE_SUCCESS

    @classmethod
    def github(cls) -> int:
        root_folder = Path(cls._args.root_folder).resolve()
        modules_folder = Config.get_modules_folder(root_folder)
        Log.set_enable_stdout(not cls._args.quiet)
        Pull.set_max_size(cls._args.max_size)

        config = Config(root_folder)

        tracks = GithubTracks(
            modules_folder=modules_folder,
            config=config,
            api_token=cls._args.token,
            after_date=parse(cls._args.after_date).date()
        )

        tracks.get_tracks(
            user_name=cls._args.user_name,
            repo_names=cls._args.repo_names,
            single=cls._args.single,
            cover=cls._args.cover,
            use_ssh=cls._args.ssh
        )

        if cls._args.clear:
            tracks.clear_tracks()

        return cls.CODE_SUCCESS

    @classmethod
    def sync(cls) -> int:
        root_folder = Path(cls._args.root_folder).resolve()
        Log.set_enable_stdout(not cls._args.quiet)
        Pull.set_max_size(cls._args.max_size)

        config = Config(root_folder)

        sync = Sync(root_folder=root_folder, config=config)
        sync.create_local_tracks()
        sync.update(
            module_ids=cls._args.module_ids,
            force=cls._args.force,
            single=cls._args.single
        )

        if cls._args.diff_file:
            markdown_text = sync.get_versions_diff()
            if markdown_text is not None:
                if isinstance(cls._args.diff_file, str):
                    diff_file = Path(cls._args.diff_file)
                    diff_file.write_text(markdown_text)

                else:
                    print(markdown_text)

        if cls._args.push:
            index = Index(root_folder=root_folder, config=config)
            index(version=cls._args.index_version, to_file=True)
            index.push_by_git(cls._args.git_branch)

        return cls.CODE_SUCCESS

    @classmethod
    def index(cls) -> int:
        root_folder = Path(cls._args.root_folder).resolve()
        Log.set_enable_stdout(False)

        config = Config(root_folder)

        index = Index(root_folder=root_folder, config=config)

        if cls._args.list:
            markdown_text = index.get_versions_table()
            print(markdown_text)

        else:
            index(version=cls._args.index_version, to_file=not cls._args.json)

            if cls._args.json:
                print_json(index.modules_json)

            elif cls._args.push:
                index.push_by_git(cls._args.git_branch)

        return cls.CODE_SUCCESS

    @classmethod
    def check(cls) -> int:
        root_folder = Path(cls._args.root_folder).resolve()
        Log.set_log_level(logging.INFO)

        if not (
            cls._args.check_id
            or cls._args.check_url
            or cls._args.remove_old
        ):
            return cls.CODE_FAILURE

        config = Config(root_folder)
        check = Check(root_folder=root_folder, config=config)

        if cls._args.check_id:
            check.ids(module_ids=cls._args.module_ids)

        if cls._args.check_url:
            check.url(module_ids=cls._args.module_ids)

        if cls._args.remove_old:
            check.old(module_ids=cls._args.module_ids)

        return cls.CODE_SUCCESS

    @classmethod
    def sitemap(cls)-> int:
        root_folder = Path(cls._args.root_folder).resolve()
        config = Config(root_folder)
        
        modules_json = cls._args.modules_json or config.get_json_folder(root_folder).joinpath("modules.json")
        base_url = cls._args.base_url
        output = cls._args.output or config.get_json_folder(root_folder).joinpath("sitemap.xml")

        Sitemap(modules_json, base_url, output)
        return cls.CODE_SUCCESS


def print_error(msg):
    print(f"Error: {msg}")


def print_json(obj: dict):
    string = json.dumps(obj, indent=2)
    print(string)


def json_parse(texts: Sequence[str], __cls: Type) -> Tuple[AttrDict, AttrDict]:
    _error = AttrDict()
    _dict = AttrDict()

    _member = AttrDict()
    for p in __cls.__mro__:
        if hasattr(p, "__annotations__"):
            _member.update(p.__annotations__)

    for text in texts:
        values = text.split("=", maxsplit=1)
        if len(values) != 2:
            continue

        key, value = values[0], values[1]

        _type = _member.get(key)
        if _type is None:
            continue

        try:
            if _type is bool:
                _dict[key] = value.lower() == "true"
            else:
                _dict[key] = _type(value)
        except BaseException as err:
            _error[key] = str(err)

    return _dict, _error



from .GitHubGraphQLAPI import GitHubGraphQLAPI
from .GitUtils import GitUtils
from .HttpUtils import HttpUtils
from .Log import Log
from .StrUtils import StrUtils
from .Sitemap import Sitemap

__all__ = [
    "GitHubGraphQLAPI",
    "GitUtils",
    "HttpUtils",
    "Log",
    "StrUtils",
    "Sitemap"
]



from enum import Enum

from .AttrDict import AttrDict
from .JsonIO import JsonIO

from .ModuleNote import ModuleNote
from .ModuleFeatures import ModuleFeatures
from .ModuleManager import ModuleManager
from .RootSolutions import RootSolutions
from .TrackOptions import TrackOptions

class TrackJson(AttrDict, JsonIO):
    id: str
    enable: bool
    verified: bool
    update_to: str
    changelog: str
    license: str
    homepage: str
    source: str
    support: str
    donate: str
    max_num: int
 
    # FoxMMM supported props
    maxApi: int
    minApi: int
    
    # MMRL supported props
    category: str
    categories: list[str]
    icon: str
    homepage: str
    donate: str
    support: str
    cover: str
    screenshots: list[str]
    license: str
    screenshots: list[str]
    readme: str
    require: list[str]
    verified: bool
    note: ModuleNote
    features: ModuleFeatures
    root: RootSolutions
    manager: ModuleManager
    
    antifeatures: list[str]
    
    options: TrackOptions

    # noinspection PyAttributeOutsideInit
    @property
    def type(self):
        if self._type is not None:
            return self._type

        if self.update_to.startswith("http"):
            if self.update_to.endswith(".json"):
                self._type = TrackType.ONLINE_JSON
            elif self.update_to.endswith(".zip"):
                self._type = TrackType.ONLINE_ZIP
            elif self.update_to.endswith(".git"):
                self._type = TrackType.GIT

        elif self.update_to.startswith("git@"):
            if self.update_to.endswith(".git"):
                self._type = TrackType.GIT

        else:
            if self.update_to.endswith(".json"):
                self._type = TrackType.LOCAL_JSON
            elif self.update_to.endswith(".zip"):
                self._type = TrackType.LOCAL_ZIP

        if self._type is None:
            self._type = TrackType.UNKNOWN

        return self._type

    def json(self):
        return AttrDict(
            type=self.type.name,
            added=self.added,
            source=self.source or None,
            antifeatures=self.antifeatures or None
        )

    def write(self, file):
        new = AttrDict()
        keys = list(self.expected_fields().keys())

        # fields without manually
        keys.extend(["added", "last_update", "versions"])

        for key in keys:
            value = self.get(key, "")
            if value is None:
                continue

            if isinstance(value, str):
                if value == "" or value.isspace():
                    continue

            new[key] = value

        JsonIO.write(new, file)

    @classmethod
    def load(cls, file):
        obj = JsonIO.load(file)
        return TrackJson(obj)

    @classmethod
    def filename(cls, yaml):
        if yaml:
            return "track.yaml"
        else:
            return "track.json"

    @classmethod
    def filenameYaml(cls):
        return "track.yaml"

    @classmethod
    def expected_fields(cls, __type=True):
        if __type:
            return cls.__annotations__

        return {k: v.__name__ for k, v in cls.__annotations__.items() if v is not None and v is not False}


class TrackType(Enum):
    UNKNOWN = 0
    ONLINE_JSON = 1
    ONLINE_ZIP = 2
    GIT = 3
    LOCAL_JSON = 4
    LOCAL_ZIP = 5



import json
import re
from datetime import datetime
from pathlib import Path
from typing import Union

import requests
from dateutil.parser import parse
from requests import HTTPError

from .StrUtils import StrUtils


class HttpUtils:
    @classmethod
    def _filter_json(cls, text: str) -> str:
        return re.sub(r",(?=\s*?[}\]])", "", text)

    @classmethod
    def load_json(cls, url: str) -> Union[list, dict]:
        response = requests.get(url, stream=True)
        if not response.ok:
            if StrUtils.is_html(response.text):
                msg = "404 not found"
            else:
                msg = response.text
            raise HTTPError(msg)

        text = cls._filter_json(response.text)
        obj = json.loads(text)

        return obj

    @classmethod
    def download(cls, url: str, out: Path) -> float:
        out.parent.mkdir(parents=True, exist_ok=True)

        response = requests.get(url, stream=True)
        if response.status_code == 200:
            block_size = 1024
            with open(out, 'wb') as file:
                for data in response.iter_content(block_size):
                    file.write(data)

            if "Last-Modified" in response.headers:
                last_modified = response.headers["Last-Modified"]
                return parse(last_modified).timestamp()
            else:
                return datetime.now().timestamp()
        else:
            out.unlink(missing_ok=True)

            if StrUtils.is_html(response.text):
                msg = "404 not found"
            else:
                msg = response.text
            raise HTTPError(msg)



import concurrent.futures
from concurrent.futures import ThreadPoolExecutor

from tabulate import tabulate

from .Config import Config
from .Pull import Pull
from ..model import (
    UpdateJson,
    TrackJson
)
from ..track import BaseTracks, LocalTracks, GithubTracks
from ..utils import Log


class Sync:
    def __init__(self, root_folder, config, tracks=None):
        self._log = Log("Sync", enable_log=config.enable_log, log_dir=config.log_dir)
        self._root_folder = root_folder
        self._pull = Pull(root_folder, config)

        self._json_folder = Config.get_json_folder(root_folder)
        self._modules_folder = Config.get_modules_folder(root_folder)
        self._config = config

        if tracks is None:
            self._tracks = BaseTracks()
        else:
            self._tracks = tracks

        self._updated_diff = list()

    def _update_jsons(self, track, force):
        module_folder = self._modules_folder.joinpath(track.id)

        if not track.enable:
            self._log.i(f"_update_jsons: [{track.id}] -> update check has been disabled")
            return None

        online_module, timestamp = self._pull.from_track(track)
        if online_module is None:
            return None

        update_json_file = module_folder.joinpath(UpdateJson.filename())
        has_yaml_track = module_folder.joinpath("track.yaml").exists()
        track_json_file = module_folder.joinpath(TrackJson.filename(has_yaml_track))
        zipfile = module_folder.joinpath(online_module.zipfile_name)
            
            
            
        if force:
            for file in module_folder.glob("*"):
                if file.name not in [
                    TrackJson.filename(has_yaml_track),
                    online_module.zipfile_name,
                    online_module.changelog_filename
                ]:
                    file.unlink()

        if update_json_file.exists():
            update_json = UpdateJson.load(update_json_file)
            update_json.update(id=track.id)
        else:
            update_json = UpdateJson(
                id=track.id,
                timestamp=timestamp,
                versions=list()
            )

        version_item = online_module.to_VersionItem(timestamp, zipfile.stat().st_size)
        update_json.versions.append(version_item)

        max_num = self._config.max_num
        if track.max_num is not None:
            max_num = track.max_num

        if len(update_json.versions) > max_num:
            old_item = update_json.versions.pop(0)
            zipfile = module_folder.joinpath(old_item.zipfile_name)
            changelog = module_folder.joinpath(old_item.changelog_filename)

            for path in [zipfile, changelog]:
                if not (path.exists() and path.is_file()):
                    continue

                self._log.d(f"_update_jsons: [{track.id}] -> remove {path.name}")
                path.unlink()

        track.last_update = timestamp
        track.versions = len(update_json.versions)

        update_json.write(update_json_file)
        track.write(track_json_file)

        if len(update_json.versions) >= 2:
            self._updated_diff.append(
                (update_json.versions[-2], online_module)
            )
        else:
            self._updated_diff.append(
                (None, online_module)
            )

        return online_module

    @staticmethod
    def _check_tracks(obj, cls):
        if type(obj) is BaseTracks:
            raise RuntimeError("tracks interface has not been created")

        return isinstance(obj, cls)

    def create_github_tracks(self, api_token, after_date=None):
        self._tracks = GithubTracks(
            modules_folder=self._modules_folder,
            config=self._config,
            api_token=api_token,
            after_date=after_date
        )
        return self._tracks

    def create_local_tracks(self):
        self._tracks = LocalTracks(
            modules_folder=self._modules_folder,
            config=self._config
        )
        return self._tracks

    def update(self, module_ids=None, force=False, single=False, **kwargs):
        user_name = kwargs.get("user_name")
        if user_name is not None:
            if self._check_tracks(self._tracks, GithubTracks):
                tracks = self._tracks.get_tracks(
                    user_name=user_name,
                    repo_names=module_ids,
                    single=single,
                    cover=kwargs.get("cover", False),
                    use_ssh=kwargs.get("use_ssh", True)
                )
            else:
                msg = f"unsupported tracks interface type [{type(self._tracks).__name__}]"
                raise RuntimeError(msg)
        else:
            tracks = self._tracks.get_tracks(module_ids)

        with ThreadPoolExecutor(max_workers=1 if single else None) as executor:
            futures = []
            for track in tracks:
                futures.append(
                    executor.submit(self._update_jsons, track=track, force=force)
                )

            for future in concurrent.futures.as_completed(futures):
                online_module = future.result()
                if online_module is not None:
                    self._log.i(f"update: [{online_module.id}] -> update to {online_module.version_display}")

    def get_versions_diff(self):
        headers = ["id", "name", "version"]
        table = []

        if len(self._updated_diff) == 0:
            return None

        for last, new in self._updated_diff:
            version = new.version_display
            if last is not None:
                version = f"{last.version_display} -> {version}"

            name = new.name.replace("|", "_")
            table.append(
                [new.id, name, version]
            )

        markdown_text = tabulate(table, headers, tablefmt="github")
        return markdown_text



from .AttrDict import AttrDict
from .JsonIO import JsonIO
from ..utils import StrUtils


class VersionItem(AttrDict):
    @property
    def id(self):
        return self.zipUrl.split("/")[-2]

    @property
    def version_display(self):
        return StrUtils.get_version_display(self.version, self.versionCode)

    @property
    def changelog_filename(self):
        return self.changelog.split("/")[-1]

    @property
    def zipfile_name(self):
        return self.zipUrl.split("/")[-1]


class UpdateJson(AttrDict, JsonIO):
    @classmethod
    def load(cls, file):
        obj = JsonIO.load(file)
        obj["versions"] = [VersionItem(_obj) for _obj in obj["versions"]]
        obj["versions"].sort(key=lambda v: v.versionCode)
        return UpdateJson(obj)

    @classmethod
    def filename(cls):
        return "update.json"



from .AttrDict import AttrDict
from .JsonIO import JsonIO
from .UpdateJson import VersionItem
from ..utils import StrUtils


class OnlineModule(AttrDict):
    @property
    def version_display(self):
        return StrUtils.get_version_display(self.version, self.versionCode)

    @property
    def changelog_filename(self):
        return StrUtils.get_filename(self.version_display, "md")

    @property
    def zipfile_name(self):
        return StrUtils.get_filename(self.version_display, "zip")

    def to_VersionItem(self, timestamp, zipfile_size):
        return VersionItem(
            timestamp=timestamp,
            version=self.version,
            versionCode=self.versionCode,
            zipUrl=self.latest.zipUrl,
            changelog=self.latest.changelog,
            size=zipfile_size
        )

    @classmethod
    def from_dict(cls, obj):
        versions = obj.get("versions")
        if versions is not None:
            obj["versions"] = [VersionItem(_obj) for _obj in versions]

        track = obj.get("track")
        if track is not None:
            obj["track"] = AttrDict(track)

        return OnlineModule(obj)


class ModulesJson(AttrDict, JsonIO):
    @property
    def size(self):
        return len(self.modules)

    def get_timestamp(self):
        value0 = self.get("timestamp")

        value1 = None
        metadata = self.get("metadata")
        if metadata is not None:
            value1 = metadata.get("timestamp")

        return value0 or value1 or 0.0

    @classmethod
    def load(cls, file):
        obj = JsonIO.load(file)
        obj["modules"] = [OnlineModule.from_dict(_obj) for _obj in obj["modules"]]
        return ModulesJson(obj)

    @classmethod
    def filename(cls):
        return "modules.json"



from .AttrDict import AttrDict
from .JsonIO import JsonIO


class ConfigJson(AttrDict, JsonIO):
    name: str
    base_url: str
    max_num: int
    enable_log: bool
    log_dir: str

    def write(self, file):
        new = AttrDict()
        for key in self.expected_fields().keys():
            value = self.get(key)
            if value is not None:
                new[key] = value

        JsonIO.write(new, file)

    @classmethod
    def load(cls, file):
        obj = JsonIO.load(file)
        return ConfigJson(obj)

    @classmethod
    def default(cls):
        return ConfigJson(
            name="Unknown",
            base_url="",
            max_num=3,
            enable_log=True,
            log_dir=None
        )

    @classmethod
    def filename(cls):
        return "config.json"

    @classmethod
    def expected_fields(cls, __type=True):
        if __type:
            return cls.__annotations__

        return {k: v.__name__ for k, v in cls.__annotations__.items()}



import zipfile

class ArchiveIO:
    def __init__(self, file, mode, compression="store"):
        self._file = file
        self._mode = mode
        
        compression_map = {
            "deflate": zipfile.ZIP_DEFLATED,
            "bzip2": zipfile.ZIP_BZIP2,
            "lzma": zipfile.ZIP_LZMA,
            "store": zipfile.ZIP_STORED
        }

        self._compression = compression_map.get(compression.lower(), zipfile.ZIP_STORED)

        try:
            self._archive = zipfile.ZipFile(file, mode, compression=self._compression)
        except Exception as e:
            raise ValueError(f"Failed to open the archive: {e}")

    def file_exists(self, name: str) -> bool:
        return name in self._archive.namelist()

    def file_read(self, name: str) -> str:
        try:
            return self._archive.read(name).decode("utf-8")
        except KeyError:
            raise FileNotFoundError(f"The file '{name}' does not exist in the archive.")

    def namelist(self) -> list:
        return self._archive.namelist()

    def close(self) -> None:
        self._archive.close()



import re
from urllib.parse import urlparse


class StrUtils:
    @classmethod
    def is_with(cls, text: str, start: str, end: str) -> bool:
        return text.startswith(start) and text.endswith(end)

    @classmethod
    def is_html(cls, text: str) -> bool:
        pattern = r'<html\s*>|<head\s*>|<body\s*>|<!doctype\s*html\s*>'
        return re.search(pattern, text, re.IGNORECASE) is not None

    @classmethod
    def is_blob_url(cls, url: str) -> bool:
        pattern = r"https://github\.com/[^/]+/[^/]+/blob/.+"
        match = re.match(pattern, url)
        if match:
            return True
        else:
            return False

    @classmethod
    def is_url(cls, text: str) -> bool:
        parse_result = urlparse(text)
        return bool(parse_result.scheme)

    @classmethod
    def get_version_display(cls, version: str, version_code: int):
        included = re.search(fr"\(.*?{version_code}.*?\)", version) is not None

        if included:
            return version
        else:
            return f"{version} ({version_code})"

    @classmethod
    def get_filename(cls, base: str, suffix: str):
        filename = base.replace(" ", "_")
        filename = re.sub(r"[^a-zA-Z0-9\-._]", "", filename)
        return f"{filename}.{suffix}"



from .AttrDict import AttrDict
from .ConfigJson import ConfigJson
from .JsonIO import JsonIO
from .LocalModule import LocalModule
from .MagiskUpdateJson import MagiskUpdateJson
from .ModulesJson import ModulesJson, OnlineModule
from .TrackJson import TrackJson, TrackType
from .UpdateJson import UpdateJson, VersionItem

__all__ = [
    "AttrDict",
    "ConfigJson",
    "JsonIO",
    "LocalModule",
    "MagiskUpdateJson",
    "ModulesJson",
    "OnlineModule",
    "TrackJson",
    "TrackType",
    "UpdateJson",
    "VersionItem",
]



from .AttrDict import AttrDict

class RootSolutions(AttrDict):
    magisk: str
    kernelsu: str
    apatch: str
    
    @classmethod
    def expected_fields(cls, __type=True):
        if __type:
            return cls.__annotations__

        return {k: v.__name__ for k, v in cls.__annotations__.items()}


import concurrent.futures
import shutil
from concurrent.futures import ThreadPoolExecutor
from datetime import date

from github import Github, Auth, UnknownObjectException
from github.Repository import Repository

from .BaseTracks import BaseTracks
from .LocalTracks import LocalTracks
from ..error import MagiskModuleError, Result
from ..model import TrackJson
from ..utils import Log, GitHubGraphQLAPI


class GithubTracks(BaseTracks):
    def __init__(self, modules_folder, config, *, api_token, after_date=None):
        self._log = Log("GithubTracks", enable_log=config.enable_log, log_dir=config.log_dir)
        self._modules_folder = modules_folder

        if after_date is None:
            after_date = date(2016, 9, 8)

        self._api_token = api_token
        self._after_date = after_date
        self._github = Github(auth=Auth.Token(api_token))
        self._graphql_api = GitHubGraphQLAPI(api_token=api_token)
        self._tracks = list()

    @Result.catching()
    def _get_from_repo_common(self, repo: Repository, use_ssh):
        if not self.is_module_repo(repo):
            raise MagiskModuleError(f"{repo.name} is not a target magisk module repository")

        try:
            update_to = repo.get_contents("update.json").download_url
            changelog = ""
        except UnknownObjectException:
            if use_ssh:
                update_to = repo.ssh_url
            else:
                update_to = repo.clone_url
            changelog = self.get_changelog(repo)

        if repo.has_issues:
            issues = f"{repo.html_url}/issues"
        else:
            issues = ""

        donate_urls = self._graphql_api.get_sponsor_url(
            owner=repo.owner.login,
            name=repo.name
        )
        if len(donate_urls) == 0:
            donate = ""
        else:
            donate = donate_urls[0]

        homepage = self._graphql_api.get_homepage_url(
            owner=repo.owner.login,
            name=repo.name
        )
        if homepage is None:
            homepage = ""

        readme = self.get_readme(repo)

        return TrackJson(
            id=repo.name,
            update_to=update_to,
            license=self.get_license(repo),
            changelog=changelog,
            homepage=homepage,
            source=repo.clone_url,
            support=issues,
            donate=donate,
            readme=readme
        )

    def _get_from_repo(self, repo, cover, use_ssh):
        self._log.d(f"_get_from_repo: repo_name = {repo.name}")

        pushed_at = self._graphql_api.get_pushed_at(
            owner=repo.owner.login,
            name=repo.name
        )
        if pushed_at is None:
            return None

        if pushed_at.date() < self._after_date:
            msg = f"pushed at {pushed_at.date()}, too old"
            self._log.w(f"_get_from_repo: [{repo.name}] -> {msg}")
            return None

        result = self._get_from_repo_common(repo, use_ssh)
        if result.is_failure:
            msg = Log.get_msg(result.error)
            self._log.e(f"_get_from_repo: [{repo.name}] -> {msg}")
            return None
        else:
            track_json: TrackJson = result.value
            LocalTracks.add_track(track_json, self._modules_folder, cover)

            return track_json

    def get_track(self, user_name, repo_name, *, cover=False, use_ssh=True):
        user = self._github.get_user(user_name)
        repo = user.get_repo(repo_name)

        return self._get_from_repo(repo, cover, use_ssh)

    def get_tracks(self, user_name, repo_names=None, *, single=False, cover=False, use_ssh=True):
        self._tracks.clear()
        self._log.i(f"get_tracks: user_name = {user_name}")

        user = self._github.get_user(user_name)
        repos = []

        if repo_names is not None:
            for repo_name in repo_names:
                try:
                    repo = user.get_repo(repo_name)
                    repos.append(repo)
                except UnknownObjectException as err:
                    msg = Log.get_msg(err)
                    self._log.e(f"get_tracks: [{repo_name}] -> {msg}")
        else:
            repos = user.get_repos()

        with ThreadPoolExecutor(max_workers=1 if single else None) as executor:
            futures = []
            for repo in repos:
                futures.append(
                    executor.submit(self._get_from_repo, repo=repo, cover=cover, use_ssh=use_ssh)
                )

            for future in concurrent.futures.as_completed(futures):
                track_json = future.result()
                if track_json is not None:
                    self._tracks.append(track_json)

        self._log.i(f"get_tracks: size = {self.size}")
        return self._tracks

    def clear_tracks(self):
        names = [track.id for track in self._tracks]
        for module_folder in self._modules_folder.glob("*/"):
            if module_folder.name not in names:
                self._log.i(f"clear_tracks: [{module_folder.name}] -> removed")
                shutil.rmtree(module_folder, ignore_errors=True)

    @property
    def size(self):
        return self._tracks.__len__()

    @property
    def tracks(self):
        return self._tracks

    @classmethod
    def get_license(cls, repo):
        try:
            _license = repo.get_license().license.spdx_id
            if _license == "NOASSERTION":
                _license = "UNKNOWN"
        except UnknownObjectException:
            _license = ""

        return _license

    @classmethod
    def get_changelog(cls, repo):
        try:
            changelog = repo.get_contents("changelog.md").download_url
        except UnknownObjectException:
            changelog = ""

        return changelog

    @classmethod
    def get_readme(cls, repo):
        try:
            readme = repo.get_contents("README.md").download_url
        except UnknownObjectException:
            readme = ""

        return readme


    @classmethod
    def is_module_repo(cls, repo):
        try:
            repo.get_contents("module.prop")
            repo.get_contents("META-INF/com/google/android/update-binary")
            repo.get_contents("META-INF/com/google/android/updater-script")
            return True
        except UnknownObjectException:
            return False



import json
import yaml
import os
import re

from sync.model import AttrDict

def represent_attr_dict(dumper: yaml.Dumper, value: AttrDict):
    return dumper.represent_mapping('tag:yaml.org,2002:map', value.items())

yaml.add_representer(AttrDict, represent_attr_dict)

class JsonIO:
    def write(self, file):
        assert isinstance(self, dict)

        _, ext = os.path.splitext(file)

        file.parent.mkdir(parents=True, exist_ok=True)

        if ext.lower() == '.yaml':
            with open(file, "w") as f:
                
                yaml.dump(self, f, indent=2, default_flow_style=False)

        else:
            with open(file, "w") as f:
                json.dump(self, f, indent=2)

    @classmethod
    def filter(cls, text):
        return re.sub(r",(?=\s*?[}\]])", "", text)

    @classmethod
    def filterArray(cls, filter, toFilter):
        return [i for i in filter if i in toFilter]
 
    @classmethod
    def load(cls, file):
        
        _, ext = os.path.splitext(file)
        
        if ext.lower() == '.yaml':
            with open(file, encoding="utf-8", mode="r") as f:
                text = cls.filter(f.read())
                obj = yaml.load(text, Loader=yaml.FullLoader)
                
                

                assert isinstance(obj, dict)
        else: 
            with open(file, encoding="utf-8", mode="r") as f:
                text = cls.filter(f.read())
                obj = json.loads(text)

                assert isinstance(obj, dict)

        return obj



from datetime import datetime

from git import Repo
from tabulate import tabulate

from .Config import Config
from ..error import Result
from ..model import (
    AttrDict,
    ModulesJson,
    UpdateJson,
    LocalModule,
    OnlineModule
)
from ..track import LocalTracks
from ..utils import Log


class Index:
    versions = [0, 1]
    latest_version = versions[-1]

    def __init__(self, root_folder, config):
        self._log = Log("Index", enable_log=config.enable_log, log_dir=config.log_dir)
        self._root_folder = root_folder

        self._modules_folder = Config.get_modules_folder(root_folder)
        self._tracks = LocalTracks(self._modules_folder, config)
        self._config = config

        self._json_folder = Config.get_json_folder(root_folder)

        # noinspection PyTypeChecker
        self.modules_json = None

    def _add_modules_json_0(self, track, update_json, online_module):
        if self.modules_json is None:
            self.modules_json = ModulesJson(
                name=self._config.name,
                timestamp=datetime.now().timestamp(),
                modules=list()
            )

        latest_item = update_json.versions[-1]

        online_module.license = track.license or ""
        online_module.states = AttrDict(
            zipUrl=latest_item.zipUrl,
            changelog=latest_item.changelog
        )

        self.modules_json.modules.append(online_module)

    def _add_modules_json_1(self, track, update_json, online_module):
        if self.modules_json is None:
            self.modules_json = ModulesJson(
                name=self._config.name,
                website=self._config.website,
                support=self._config.support,
                donate=self._config.donate,
                submission=self._config.submission,
                cover=self._config.cover,
                description=self._config.description,
                metadata=AttrDict(
                    version=1,
                    timestamp=datetime.now().timestamp()
                ),
                modules=list()
            )

        online_module.track = track.json()
        online_module.versions = update_json.versions

        self.modules_json.modules.append(online_module)

    def _add_modules_json(self, track, update_json, online_module, version):
        if version not in self.versions:
            raise RuntimeError(f"unsupported version: {version}")

        func = getattr(self, f"_add_modules_json_{version}")
        func(
            track=track,
            update_json=update_json,
            online_module=online_module,
        )

    def get_online_module(self, track, zip_file):
        @Result.catching()
        def get_online_module():
            local_module = LocalModule.load(zip_file, track, self._config)
            return OnlineModule.from_dict(local_module)

        result = get_online_module()
        if result.is_failure:
            msg = Log.get_msg(result.error)
            self._log.e(f"get_online_module: [{track.id}] -> {msg}")
            return None

        else:
            return result.value

    def __call__(self, version, to_file):
        for track in self._tracks.get_tracks():
            module_folder = self._modules_folder.joinpath(track.id)
            update_json_file = module_folder.joinpath(UpdateJson.filename())
            if not update_json_file.exists():
                continue

            update_json = UpdateJson.load(update_json_file)
            latest_item = update_json.versions[-1]

            zip_file = module_folder.joinpath(latest_item.zipfile_name)
            if not zip_file.exists():
                continue

            online_module = self.get_online_module(track, zip_file)
            if online_module is None:
                continue

            self._add_modules_json(
                track=track,
                update_json=update_json,
                online_module=online_module,
                version=version
            )

        self.modules_json.modules.sort(key=lambda v: v.id)
        if to_file:
            json_file = self._json_folder.joinpath(ModulesJson.filename())
            self.modules_json.write(json_file)

        return self.modules_json

    def push_by_git(self, branch):
        json_file = self._json_folder.joinpath(ModulesJson.filename())
        timestamp = ModulesJson.load(json_file).get_timestamp()
        msg = f"Update by CLI ({datetime.fromtimestamp(timestamp)})"

        repo = Repo(self._root_folder)
        repo.git.add(all=True)
        repo.index.commit(msg)
        repo.remote().push(branch)

    def get_versions_table(self):
        headers = ["id", "name", "latest version"]
        table = []

        for track in self._tracks.get_tracks():
            module_folder = self._modules_folder.joinpath(track.id)
            update_json_file = module_folder.joinpath(UpdateJson.filename())

            if not update_json_file.exists():
                table.append(
                    [track.id, "-", "-"]
                )
                continue

            update_json = UpdateJson.load(update_json_file)
            latest = update_json.versions[-1]
            zip_file = module_folder.joinpath(latest.zipfile_name)
            online_module = self.get_online_module(track, zip_file)

            if online_module is not None:
                name = online_module.name.replace("|", "-")
                table.append(
                    [online_module.id, name, online_module.version_display]
                )
            else:
                table.append(
                    [track.id, "-", "-"]
                )

        markdown_text = tabulate(table, headers, tablefmt="github")
        return markdown_text



import json
import validators
import xml.etree.ElementTree as ET

from xml.dom import minidom
from datetime import datetime

def Sitemap(modules_json, base_url, output):
    # Open modules.json
    with open(modules_json) as json_data:
        data = json.load(json_data)

    # Access the "modules" object
    modules = data.get("modules", [])

    # Create the root element for the sitemap
    urlset = ET.Element("urlset", xmlns="http://www.sitemaps.org/schemas/sitemap/0.9", **{"xmlns:image": "http://www.google.com/schemas/sitemap-image/1.1"})

    # Add URLs to the sitemap
    for module in modules:
        # Convert Unix timestamp to datetime object
        dt_object = datetime.utcfromtimestamp(module["timestamp"])

        # Format the datetime object as 'YYYY-MM-DD'
        formatted_date = dt_object.strftime('%Y-%m-%d')

        module_id = module["id"]
        url = base_url + module_id
        url_element = ET.SubElement(urlset, "url")
        loc_element = ET.SubElement(url_element, "loc")
        lastmod_element = ET.SubElement(url_element, "lastmod")
        
        cover = module.get("cover") if module.get("cover") else module.get("track").get("cover") 
        if cover and validators.url(cover):
            url_image_image = ET.SubElement(url_element, "image:image")
            url_image_image_loc = ET.SubElement(url_image_image, "image:loc")
            url_image_image_loc.text = cover

        lastmod_element.text = formatted_date
        loc_element.text = url

    # Convert the ElementTree to a string
    sitemap = ET.tostring(urlset, encoding="utf-8", method="xml")

    xmlstr = minidom.parseString(sitemap).toprettyxml(indent="   ")

    # Write to a file
    with open(output, "w") as f:
        f.write(xmlstr)

    print("Sitemap generated successfully.")


from setuptools import setup, find_packages

from sync import __version__

setup(
    name="mmrl-util",
    version=__version__.get_version(),
    description="A tool to build your own Magisk Modules Repository",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Der_Googler & Sanmer",
    author_email="jimmy@dergooglr.com",
    url="https://github.com/Googlers-Repo/magisk-modules-repo-util",
    license="GPL-3.0",
    packages=find_packages(),
    install_requires=[
        "pygithub>=1.59.0",
        "python-dateutil>=2.8.2",
        "requests>=2.31.0",
        "tabulate>=0.9.0",
        "gitpython>=3.1.37",
        "validators>=0.28.3",
        "pyyaml>=6.0.2",
        "python-magic>=0.4.27"
    ],
    entry_points={
        "console_scripts": [
            "mmrl-util=sync.__main__:main",
        ],
    },
    python_requires=">=3.7",
)


import functools
import os
import shutil
from pathlib import Path
from typing import Optional

from git import Repo, InvalidGitRepositoryError, GitCommandError


class GitUtils:
    @classmethod
    @functools.lru_cache()
    def current_branch(cls, repo_dir: Path) -> Optional[str]:
        try:
            return Repo(repo_dir).active_branch.name
        except InvalidGitRepositoryError:
            return None

    @classmethod
    def clone_and_zip(cls, url: str, out: Path) -> float:
        repo_dir = out.with_suffix("")
        if repo_dir.exists():
            shutil.rmtree(repo_dir)

        try:
            repo = Repo.clone_from(url, repo_dir)
            last_committed = float(repo.commit().committed_date)
        except GitCommandError:
            shutil.rmtree(repo_dir, ignore_errors=True)
            raise GitCommandError(f"clone failed: {url}")

        for path in repo_dir.iterdir():
            if path.name.startswith(".git"):
                if path.is_dir():
                    shutil.rmtree(path, ignore_errors=True)
                if path.is_file():
                    path.unlink(missing_ok=True)

                continue

            os.utime(path, (last_committed, last_committed))

        try:
            shutil.make_archive(repo_dir.as_posix(), format="zip", root_dir=repo_dir)
            shutil.rmtree(repo_dir)
        except FileNotFoundError:
            raise FileNotFoundError(f"archive failed: {repo_dir.as_posix()}")

        return last_committed



import shutil
from datetime import datetime
from typing import List

from tabulate import tabulate

from .BaseTracks import BaseTracks
from ..error import Result
from ..model import TrackJson
from ..utils import Log


class LocalTracks(BaseTracks):
    def __init__(self, modules_folder, config):
        self._log = Log("LocalTracks", enable_log=config.enable_log, log_dir=config.log_dir)
        self._modules_folder = modules_folder

        self._tracks: List[TrackJson] = list()

    @Result.catching()
    def _get_from_file(self, file):
        return TrackJson.load(file)

    def get_track(self, module_id):
        module_folder = self._modules_folder.joinpath(module_id)
        has_yaml_track = module_folder.joinpath("track.yaml").exists()
        json_file = module_folder.joinpath(TrackJson.filename(has_yaml_track))

        result = self._get_from_file(json_file)
        if result.is_failure:
            msg = Log.get_msg(result.error)
            self._log.e(f"get_track: [{module_id}] -> {msg}")

            return None
        else:
            return result.value

    def get_tracks(self, module_ids=None):
        self._tracks.clear()
        self._log.i(f"get_tracks: modules_folder = {self._modules_folder}")

        if module_ids is None:
            module_ids = [_dir.name for _dir in sorted(self._modules_folder.glob("[!.]*/"))]

        for module_id in module_ids:
            track_json = self.get_track(module_id)
            if track_json is not None:
                self._tracks.append(track_json)

        self._log.i(f"get_tracks: size = {self.size}")
        return self._tracks

    def get_tracks_table(self):
        if self.size == 0:
            self.get_tracks()

        headers = ["id", "add time", "last update", "versions"]
        table = []

        for track in self.tracks:
            if track.versions is None:
                last_update = "-"
                versions = 0
            else:
                last_update = datetime.fromtimestamp(track.last_update).date()
                versions = track.versions

            table.append([
                track.id,
                datetime.fromtimestamp(track.added).date(),
                last_update,
                versions
            ])

        markdown_text = tabulate(table, headers, tablefmt="github")
        return markdown_text

    @property
    def size(self):
        return self._tracks.__len__()

    @property
    def tracks(self):
        return self._tracks

    @classmethod
    def add_track(cls, track, modules_folder, cover=True):
        module_folder = modules_folder.joinpath(track.id)
        has_yaml_track = module_folder.joinpath("track.yaml").exists()
        json_file = module_folder.joinpath(TrackJson.filename(has_yaml_track))

        if not json_file.exists():
            track.added = datetime.now().timestamp()
            track.enable = True
            track.verified = False
            track.write(json_file)
        elif cover:
            old = TrackJson.load(json_file)
            old.added = old.added or datetime.fromtimestamp(json_file.stat().st_ctime)
            old.enable = old.enable or True
            old.verified = old.verified or False

            old.update(track)
            old.write(json_file)

    @classmethod
    def del_track(cls, module_id, modules_folder):
        module_folder = modules_folder.joinpath(module_id)
        shutil.rmtree(module_folder, ignore_errors=True)

    @classmethod
    def update_track(cls, track, modules_folder):
        module_folder = modules_folder.joinpath(track.id)
        has_yaml_track = module_folder.joinpath("track.yaml").exists()
        json_file = module_folder.joinpath(TrackJson.filename(has_yaml_track))

        if not json_file.exists():
            return

        old = TrackJson.load(json_file)
        old.update(track)
        old.write(json_file)



from pathlib import Path

from .AttrDict import AttrDict
from .JsonIO import JsonIO
from ..utils import HttpUtils, StrUtils


class MagiskUpdateJson(AttrDict):
    @property
    def version_display(self):
        return StrUtils.get_version_display(self.version, self.versionCode)

    @property
    def zipfile_name(self):
        return StrUtils.get_filename(self.version_display, "zip")

    @classmethod
    def load(cls, path):
        if isinstance(path, str):
            obj = HttpUtils.load_json(path)
        elif isinstance(path, Path):
            obj = JsonIO.load(path)
        else:
            raise ValueError(f"unsupported type {type(path).__name__}")

        try:
            obj["versionCode"] = int(obj["versionCode"])
        except ValueError:
            msg = f"wrong type of versionCode, expected int but got {type(obj['versionCode']).__name__}"
            raise ValueError(msg)
        except TypeError:
            raise ValueError("versionCode does not exist in module.prop")

        return MagiskUpdateJson(obj)
